/* -*- c-file-style: "linux"; -*- */
/**********************************************************************
 * iproxy.c                                                   July 2001
 * Horms                                             horms@vergenet.net
 * Andrew Tridgell                                     tridge@samba.org
 *
 * iproxy
 * UDP - TCP proxy
 * Copyright (C) 2001  Andrew Tridgell <tridge@samba.org>
 *                     Horms <horms@vergenet.net>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 * 02111-1307  USA
 *
 **********************************************************************/

#define _USE_BSD
#include <sys/types.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <time.h>
#include <signal.h>

#include "util.h"
#include "iproxy.h"
#include "packet.h"
#include "interface.h"
#include "connection.h"

/*#define IPROXY_PAUSE 500000000*/
#define IPROXY_PAUSE 0

#define KEEPALIVE_TIME (time_t)5 

static uint16 iproxy_port = PROXY_PORT;
static uint16 listen_port = PROXY_PORT;
static uint32 iproxy_sid = PROXY_SID;
static uint32 iproxy_cid = PROXY_CID;
uint32 mcast_ttl = TTL_MULTICAST;
static char *xcast_ip = ADDR_BROADCAST;
static pid_t iproxy_pid;
uint32 iproxy_mode = MODE_BROADCAST|MODE_CLIENT;
int iproxy_verbose;

static struct connection con;


static void do_proxy(int sock)
{
	ssize_t n = 0;
	int fd_in = 0;
	uint32 offset = 0;
	socklen_t socklen;
        fd_set read_template;
	struct in_addr addr;
	struct packet packet;
	struct sockaddr saddr;
	struct sockaddr_in iaddr;

        socklen=sizeof(saddr);

	if (iproxy_sid != IPROXY_LIST_SID) {
	        listen(sock, 10);
	        while (1) {
		        fd_in = accept(sock, &saddr, &socklen);
                        if(fd_in < 0){
                                perror("do_proxy: accept");
                                continue;
                        }
		        if (fork() == 0) break;
		        close(fd_in);
	        }
	        close(sock);
        }

        iproxy_pid = getpid();
        iproxy_cid = iproxy_pid;         /* Bad Horms */
	
        /* We can just pass the mcast_ip as this
         * argument isn't used unless we are in multicast mode
         */
        if(interface_open_fd(ADDR_ANY, xcast_ip, 0)){
                fprintf(stderr, "do_proxy: interface_open_fd\n");
                exit(-1);
        }

        if (inet_pton(AF_INET, xcast_ip, &addr) <= 0) {
                perror("inet_pton");
                exit(-1);
        }

	bzero(&iaddr, sizeof(iaddr));
	iaddr.sin_port = htons(iproxy_port);
	iaddr.sin_family = AF_INET;
	iaddr.sin_addr = addr;
        
        connection_initialise(&con, fd_in, &iaddr, &addr, iproxy_cid, 
                        iproxy_sid);

	/* this initial packet is used to prime the connection */
        packet_send_primer(&iaddr, iproxy_cid, iproxy_sid);

        while(1){
                struct timeval tv;
                int fd;

                FD_ZERO(&read_template);
                if(!interface_add_to_read_template(&read_template))
                        break;
                if(iproxy_sid != IPROXY_LIST_SID) {
                        connection_read_template(&con, &read_template);
                }
                
                tv.tv_sec = KEEPALIVE_TIME;
                tv.tv_usec = 0;
                
                n=select(FD_SETSIZE, &read_template, NULL, NULL, &tv);

                if (n == 0) {
                        uint32 status;

                        /* timeout - send a keepalive */
                        if(iproxy_sid == IPROXY_LIST_SID){
                                printf("[%d] Tired of waiting for Hello "
                                                "replies\n", iproxy_pid);
                                exit(0);
                        }
                        status=connection_get_status(&con);
                        if(status==STATUS_PURGED || status==STATUS_CLOSED)
                                continue;

                        packet_send_keepalive(&iaddr, iproxy_cid, iproxy_sid);
                        continue;
                }

                if(n<0){
                        if(errno!=EINTR){
                                perror("do_proxy: select");
                                exit(-1);
                        }
                        continue; /* Ignore EINTR */
                }
                else if(n==0){
                        fprintf(stderr, "do_proxy: select returned 0");
                        exit(-1);
                }
                else if(FD_ISSET(fd_in, &read_template)){
                        if(connection_read_tcp(&con))
                                exit(0);
                }

                interface_set_fd_start();
                while((fd=interface_set_fd_next(&read_template))>=0){
                        bzero(&packet, sizeof(packet));
                        if((n=read(fd, &packet, sizeof(packet)))<=0) {
                                break;
                        }
                        if (iproxy_verbose > 0) {
                                printf("[%d] received %d byte reply\n", 
                                                iproxy_pid, n);
                        }
                        
                        if (packet.version != htonl(IPROXY_PROTO_VERSION)){
                                printf("[%d] packet for unknown iproxy "
                                                "version\n", iproxy_pid);
                                continue;
                        }

                        if (ntohl(packet.length) != n ||
                                        ntohl(packet.length) < PACKET_HDR_LEN){
                                printf("[%d] length missmatch\n", iproxy_pid);
                                continue;
                        }

                        if (ntohl(packet.cid) != iproxy_cid){
                                printf("[%d] cid missmatch %d!=%d\n", 
                                                iproxy_pid, ntohl(packet.cid), 
                                          iproxy_cid);
                                continue;
                        }
 
                        if (ntohl(packet.sid) == IPROXY_LIST_SID){
                                if(ntohl(packet.length) != PACKET_HDR_LEN){
                                        printf("[%d] %s", iproxy_pid, 
                                                        packet.buf);
                                        fflush(stdout);
                                }
                                continue;
                        }

                        if (ntohl(packet.sid) != iproxy_sid){
                                printf("[%d] sid missmatch %d!=%d\n", 
                                                iproxy_pid, ntohl(packet.sid), 
                                                iproxy_sid);
                                continue;
                        }

                        if(htonl(packet.checksum) != packet_checksum(&packet)){
                                printf("[%d] checksum missmatch 0x%x!=0x%x\n",
                                                iproxy_pid,
                                                ntohl(packet.checksum),
                                                packet_checksum(&packet));
                                continue;
                        }

                        if (ntohl(packet.length) == PACKET_HDR_LEN){
                                int status;
                                status=connection_command_packet(&con, 
                                                &packet, IPROXY_PAUSE);
                                if(status != 1)
                                        continue;

                                if(status < 0 ){
                                        fprintf(stderr, "do_proxy: "
                                                "connection_command_packet\n");
                                        if(status == -1)
                                                continue;
                                }
                                
                                exit(0);
                          } 

                          if (ntohl(packet.offset) != offset) {
                                  printf("[%d] out of order packet! %d!=%d\n", 
                                                  iproxy_pid,
                                                  ntohl(packet.offset), 
                                                  offset);
                                  /* Ack in case it is a resend */
                                  packet_send_ack(&iaddr, iproxy_cid,
                                                  iproxy_sid);
                                  continue;
                          }

                          write_all(fd_in, packet.buf, ntohl(packet.length) - 
                                          PACKET_HDR_LEN);

                          offset += ntohl(packet.length)-PACKET_HDR_LEN;
                          
                          packet_send_ack(&iaddr, iproxy_cid, iproxy_sid);
                }
                interface_set_fd_end();
        }

        exit(0);
}

static void usage(exit_status)
{
	fprintf(exit_status?stderr:stdout, "
iproxy-client [options]
    -h            help
    -l            list servers
    -p PORT       listen port number (TCP)
    -d PORT       dest port (UDP)
    -c CID        client id number 
    -s SID        server id number
    -I IP         IP to send packets to
    -T TTL        ttl for multicast packets
    -b            Use broadicast (default)
    -m            Use multicast
    -u            Use unicast
    -v            increase verbosity
");

	exit(exit_status);
}

static void reaper(int sig){
  int status;
  
  signal(SIGCHLD, (void(*)(int))reaper);
  while(wait3(&status, WNOHANG, 0)>0){
    ;
  }
}


static void dumper(int sig){
  char *tmp_str;
  time_t time_tmp;
  
  signal(sig, (void(*)(int))dumper);

  if(connection_get_status(&con) == STATUS_PURGED)
          return;

  time_tmp=time(NULL);
  tmp_str=ctime(&time_tmp);
  *(tmp_str+strlen(tmp_str)-1)='\0';
  printf("[%d] Dumping connection table %s:\n", iproxy_pid, tmp_str);
  connection_dump(&con);
}


int main(int argc, char *argv[])
{
	int fd;
	int c;
        int addr_set=0;

	while ((c = getopt(argc, argv, "c:s:I:T:bmulp:d:vh")) != -1) {
		switch (c) {
		case 'c':
			iproxy_cid = atoi(optarg);
			break;
		case 's':
			iproxy_sid = atoi(optarg);
			break;
		case 'I':
			xcast_ip = optarg;
                        addr_set++;
			break;
		case 'T':
			mcast_ttl = atoi(optarg);
			break;
                case 'b':
                        iproxy_mode = MODE_BROADCAST|MODE_CLIENT;
                        if(!addr_set)
                                xcast_ip = ADDR_BROADCAST;
                        break;
                case 'm':
                        iproxy_mode = MODE_MULTICAST|MODE_CLIENT;
                        if(!addr_set)
                                xcast_ip = ADDR_MULTICAST;
                        break;
                case 'u':
                        iproxy_mode = MODE_UNICAST|MODE_CLIENT;
                        if(!addr_set)
                                xcast_ip = ADDR_UNICAST;
                        break;
		case 'l':
			iproxy_sid = IPROXY_LIST_SID;
			break;
		case 'p':
			listen_port = atoi(optarg);
			break;
		case 'd':
			iproxy_port = atoi(optarg);
			break;
		case 'v':
			iproxy_verbose++;
			break;
		case 'h':
                        usage(0);
                        break;
		default:
			usage(-1);
			break;
		}
		
	}

        iproxy_pid = getpid();

        if(xcast_ip==NULL) {
                fprintf(stderr, "Need to set an address using -I\n");
                usage(-1);
        }

	fd = open_socket_in(SOCK_STREAM, ADDR_ANY, listen_port);
	if (fd == -1) {
		fprintf(stderr,"Failed to listen on %d\n", listen_port);
		exit(-1);
	}

        if(iproxy_mode&MODE_MULTICAST){
                struct in_addr addr;

                if (inet_pton(AF_INET, xcast_ip, &addr) <= 0) {
                        perror("inet_pton");
                        exit(-1);
                }
                if(!IN_MULTICAST(ntohl(addr.s_addr))) {
                        printf("not a multicast address: %s\n", 
                               inet_ntoa(addr));
                        exit(-1);
                }
        }

        if( !(iproxy_mode&MODE_UNICAST) && interfaces_get()<0){
                exit(1);
        }
        
        signal(SIGCHLD, (void(*)(int))reaper);
        signal(SIGUSR1, (void(*)(int))dumper);

	do_proxy(fd);
	return 0;
}
